import datetime
import itertools
from typing import Optional, Tuple, Union, List

import pandas as pd

SUPPORTED_BEGIN_YEAR = 2015


class SearchDates:
    strings: List[str]

    def __init__(
        self,
        start_date: str,
        end_date: Optional[str] = None
    ):
        if end_date is None:
            end_date = start_date
        self.start_date = start_date
        self.end_date = end_date
        self.strings = self.make_strs()

    def make_strs(self) -> List[str]:
        dates = pd.date_range(start=self.start_date, end=self.end_date, freq='h')
        date_strs = [_date_to_str(date) for date in dates]
        return date_strs


def _date_to_str(date: pd.Timestamp):
    mdy = date.strftime("%Y-%m-%d")
    if date.hour == 0:
        h = '0'
    else:
        h = date.strftime("%H").lstrip("0")
    return f'{mdy}-{h}'


def _zero_pad(num: int) -> str:
    num_str = str(num)
    if len(num_str) > 1:
        return num_str
    elif len(num_str) == 1:
        return f"0{num}"
    else:
        raise ValueError(f"invalid num: {num}")



import datetime
from typing import Optional, Union, Tuple, Sequence

import requests
import gzip
import json

from gharchive.exc import NoArchiveForDateException, NoArchiveMatchingCriteraException
from gharchive.logger import logger
from gharchive.models import Archive
from gharchive.search import SearchDates

BASE_URL = "https://data.gharchive.org/"


class GHArchive:
    """
    Main class for getting Github Archive data.

    :Examples:

        >>> from gharchive import GHArchive
        >>> gh = GHArchive()
        ...
        >>> data = gh.get('6/8/2020', '6/10/2020', filters=[
        >>>     ('repo.name', 'bitcoin/bitcoin'),
        >>>     ('type', 'WatchEvent')
        >>> ])
    """

    def get(
        self,
        start_date: str,
        end_date: Optional[str] = None,
        filters: Optional[Sequence[Tuple[str, Union[int, float, str]]]] = None,
    ) -> Archive:
        """
        Get data from the Github Archive

        :param start_date: string of date or datetime to start collection
        :param end_date: string of date or datetime to end collection, defaults to
        the same as start date
        :param filters: filters in the format of two-element tuples for which
        the first element is the . separated path to look up a value from the
        Archive object and the second is a value which it should be equal to
        :return: Archive object containing Github Archive data

        :Examples:

                >>> from gharchive import GHArchive
                >>> gh = GHArchive()
                ...
                >>> data = gh.get('6/8/2020', '6/10/2020', filters=[
                >>>     ('repo.name', 'bitcoin/bitcoin'),
                >>>     ('type', 'WatchEvent')
                >>> ])
        """
        dates = SearchDates(start_date, end_date)
        archive = self._get_many_and_decode(dates, filters)
        return archive

    def _get_many_and_decode(
        self,
        dates: SearchDates,
        filters: Optional[Sequence[Tuple[str, Union[int, float, str]]]] = None,
    ) -> Archive:
        full_archive = Archive([])
        for search_date in dates.strings:
            try:
                resp = self._get(str(search_date))
            except NoArchiveForDateException as e:
                logger.warning(str(e))
                continue
            archive = self._decode_response(resp, filters)
            full_archive += archive

        if full_archive is None:
            raise NoArchiveMatchingCriteraException(
                f"Could not find any archives with search dates: {dates.strings}"
            )

        return full_archive

    def _get(self, date: str) -> requests.Response:
        url = f"{BASE_URL}{date}.json.gz"
        logger.debug(f"Requesting url: {url}")
        resp = requests.get(url)
        if resp.status_code != 200:
            raise NoArchiveForDateException(
                f"Got status code {resp.status_code} for date {date}. Content: {resp.text}"
            )
        return resp

    def _decode_response(
        self,
        resp: requests.Response,
        filters: Optional[Sequence[Tuple[str, Union[int, float, str]]]] = None,
    ) -> Archive:
        arch = Archive.from_response(resp)
        if filters:
            arch = arch.filter(filters)
        return arch



import datetime

from gharchive.search import SearchDates


def test_create_with_begin_only():
    date = SearchDates(datetime.datetime(2020, 1, 1, 5))
    assert date.strings == ["2020-01-01-5"]


def test_create_with_begin_and_end():
    date = SearchDates(
        datetime.datetime(2020, 1, 1, 5), datetime.datetime(2020, 1, 1, 10),
    )
    assert date.strings == [
        "2020-01-01-5",
        "2020-01-01-6",
        "2020-01-01-7",
        "2020-01-01-8",
        "2020-01-01-9",
        "2020-01-01-10",
    ]



import logging
import sys

logger = logging.getLogger('gharchive')
logger.setLevel(logging.INFO)
formatter = logging.Formatter('%(asctime)s: %(name)s - %(filename)s L%(lineno)s - %(levelname)s - %(message)s')

ch = logging.StreamHandler(stream=sys.stdout)
ch.setLevel(logging.DEBUG)
ch.setFormatter(formatter)
logger.addHandler(ch)


import gzip
import json
from typing import (
    Optional,
    Any,
    List,
    TypeVar,
    Type,
    cast,
    Callable,
    Sequence,
    Tuple,
    Union,
)
from datetime import datetime
import dateutil.parser
import requests

T = TypeVar("T")


def from_int(x: Any) -> int:
    assert isinstance(x, int) and not isinstance(x, bool)
    return x


def from_none(x: Any) -> Any:
    assert x is None
    return x


def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def from_bool(x: Any) -> bool:
    assert isinstance(x, bool)
    return x


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


def from_list(f: Callable[[Any], T], x: Any) -> List[T]:
    assert isinstance(x, list)
    return [f(y) for y in x]


def from_datetime(x: Any) -> datetime:
    return dateutil.parser.parse(x)


# TODO: support Timeline API format
#
# Any records from 2/12/2011-12/31/2014 were from the deprecated Timeline API
# and so come in a different format. Need to parse them into the same models.
# Currently the code will fail to parse any responses from this time period.
# There is already a 2012-06-14-15.json.gz in the test data which triggers this.
# Also update gharchive.search.SUPPORTED_BEGIN_YEAR after this and add constraints
# for not being earlier in 2011.


class Actor:
    id: Optional[int]
    login: Optional[str]
    display_login: Optional[str]
    gravatar_id: Optional[str]
    url: Optional[str]
    avatar_url: Optional[str]

    def __init__(
        self,
        id: Optional[int],
        login: Optional[str],
        display_login: Optional[str],
        gravatar_id: Optional[str],
        url: Optional[str],
        avatar_url: Optional[str],
    ) -> None:
        self.id = id
        self.login = login
        self.display_login = display_login
        self.gravatar_id = gravatar_id
        self.url = url
        self.avatar_url = avatar_url

    @staticmethod
    def from_dict(obj: Any) -> "Actor":
        assert isinstance(obj, dict)
        id = from_union([from_int, from_none], obj.get("id"))
        login = from_union([from_str, from_none], obj.get("login"))
        display_login = from_union([from_str, from_none], obj.get("display_login"))
        gravatar_id = from_union([from_str, from_none], obj.get("gravatar_id"))
        url = from_union([from_str, from_none], obj.get("url"))
        avatar_url = from_union([from_str, from_none], obj.get("avatar_url"))
        return Actor(id, login, display_login, gravatar_id, url, avatar_url)

    def to_dict(self) -> dict:
        result: dict = {}
        result["id"] = from_union([from_int, from_none], self.id)
        result["login"] = from_union([from_str, from_none], self.login)
        result["display_login"] = from_union([from_str, from_none], self.display_login)
        result["gravatar_id"] = from_union([from_str, from_none], self.gravatar_id)
        result["url"] = from_union([from_str, from_none], self.url)
        result["avatar_url"] = from_union([from_str, from_none], self.avatar_url)
        return result


class Author:
    name: Optional[str]
    email: Optional[str]

    def __init__(self, name: Optional[str], email: Optional[str]) -> None:
        self.name = name
        self.email = email

    @staticmethod
    def from_dict(obj: Any) -> "Author":
        assert isinstance(obj, dict)
        name = from_union([from_str, from_none], obj.get("name"))
        email = from_union([from_str, from_none], obj.get("email"))
        return Author(name, email)

    def to_dict(self) -> dict:
        result: dict = {}
        result["name"] = from_union([from_str, from_none], self.name)
        result["email"] = from_union([from_str, from_none], self.email)
        return result


class Commit:
    sha: Optional[str]
    author: Optional[Author]
    message: Optional[str]
    distinct: Optional[bool]
    url: Optional[str]

    def __init__(
        self,
        sha: Optional[str],
        author: Optional[Author],
        message: Optional[str],
        distinct: Optional[bool],
        url: Optional[str],
    ) -> None:
        self.sha = sha
        self.author = author
        self.message = message
        self.distinct = distinct
        self.url = url

    @staticmethod
    def from_dict(obj: Any) -> "Commit":
        assert isinstance(obj, dict)
        sha = from_union([from_str, from_none], obj.get("sha"))
        author = from_union([Author.from_dict, from_none], obj.get("author"))
        message = from_union([from_str, from_none], obj.get("message"))
        distinct = from_union([from_bool, from_none], obj.get("distinct"))
        url = from_union([from_str, from_none], obj.get("url"))
        return Commit(sha, author, message, distinct, url)

    def to_dict(self) -> dict:
        result: dict = {}
        result["sha"] = from_union([from_str, from_none], self.sha)
        result["author"] = from_union(
            [lambda x: to_class(Author, x), from_none], self.author
        )
        result["message"] = from_union([from_str, from_none], self.message)
        result["distinct"] = from_union([from_bool, from_none], self.distinct)
        result["url"] = from_union([from_str, from_none], self.url)
        return result


class Payload:
    ref: Optional[str]
    ref_type: Optional[str]
    pusher_type: Optional[str]
    push_id: Optional[int]
    size: Optional[int]
    distinct_size: Optional[int]
    head: Optional[str]
    before: Optional[str]
    commits: Optional[List[Commit]]

    def __init__(
        self,
        ref: Optional[str],
        ref_type: Optional[str],
        pusher_type: Optional[str],
        push_id: Optional[int],
        size: Optional[int],
        distinct_size: Optional[int],
        head: Optional[str],
        before: Optional[str],
        commits: Optional[List[Commit]],
    ) -> None:
        self.ref = ref
        self.ref_type = ref_type
        self.pusher_type = pusher_type
        self.push_id = push_id
        self.size = size
        self.distinct_size = distinct_size
        self.head = head
        self.before = before
        self.commits = commits

    @staticmethod
    def from_dict(obj: Any) -> "Payload":
        assert isinstance(obj, dict)
        ref = from_union([from_str, from_none], obj.get("ref"))
        ref_type = from_union([from_str, from_none], obj.get("ref_type"))
        pusher_type = from_union([from_str, from_none], obj.get("pusher_type"))
        push_id = from_union([from_int, from_none], obj.get("push_id"))
        size = from_union([from_int, from_none], obj.get("size"))
        distinct_size = from_union([from_int, from_none], obj.get("distinct_size"))
        head = from_union([from_str, from_none], obj.get("head"))
        before = from_union([from_str, from_none], obj.get("before"))
        commits = from_union(
            [lambda x: from_list(Commit.from_dict, x), from_none], obj.get("commits")
        )
        return Payload(
            ref,
            ref_type,
            pusher_type,
            push_id,
            size,
            distinct_size,
            head,
            before,
            commits,
        )

    def to_dict(self) -> dict:
        result: dict = {}
        result["ref"] = from_union([from_str, from_none], self.ref)
        result["ref_type"] = from_union([from_str, from_none], self.ref_type)
        result["pusher_type"] = from_union([from_str, from_none], self.pusher_type)
        result["push_id"] = from_union([from_int, from_none], self.push_id)
        result["size"] = from_union([from_int, from_none], self.size)
        result["distinct_size"] = from_union([from_int, from_none], self.distinct_size)
        result["head"] = from_union([from_str, from_none], self.head)
        result["before"] = from_union([from_str, from_none], self.before)
        result["commits"] = from_union(
            [lambda x: from_list(lambda x: to_class(Commit, x), x), from_none],
            self.commits,
        )
        return result


class Repo:
    id: Optional[int]
    name: Optional[str]
    url: Optional[str]

    def __init__(
        self, id: Optional[int], name: Optional[str], url: Optional[str]
    ) -> None:
        self.id = id
        self.name = name
        self.url = url

    @staticmethod
    def from_dict(obj: Any) -> "Repo":
        assert isinstance(obj, dict)
        id = from_union([from_int, from_none], obj.get("id"))
        name = from_union([from_str, from_none], obj.get("name"))
        url = from_union([from_str, from_none], obj.get("url"))
        return Repo(id, name, url)

    def to_dict(self) -> dict:
        result: dict = {}
        result["id"] = from_union([from_int, from_none], self.id)
        result["name"] = from_union([from_str, from_none], self.name)
        result["url"] = from_union([from_str, from_none], self.url)
        return result


class ArchiveElement:
    id: Optional[str]
    type: Optional[str]
    actor: Optional[Actor]
    repo: Optional[Repo]
    payload: Optional[Payload]
    public: Optional[bool]
    created_at: Optional[datetime]

    def __init__(
        self,
        id: Optional[str],
        type: Optional[str],
        actor: Optional[Actor],
        repo: Optional[Repo],
        payload: Optional[Payload],
        public: Optional[bool],
        created_at: Optional[datetime],
    ) -> None:
        self.id = id
        self.type = type
        self.actor = actor
        self.repo = repo
        self.payload = payload
        self.public = public
        self.created_at = created_at

    @staticmethod
    def from_dict(obj: Any) -> "ArchiveElement":
        assert isinstance(obj, dict)
        id = from_union([from_str, from_none], obj.get("id"))
        type = from_union([from_str, from_none], obj.get("type"))
        actor = from_union([Actor.from_dict, from_none], obj.get("actor"))
        repo = from_union([Repo.from_dict, from_none], obj.get("repo"))
        payload = from_union([Payload.from_dict, from_none], obj.get("payload"))
        public = from_union([from_bool, from_none], obj.get("public"))
        created_at = from_union([from_datetime, from_none], obj.get("created_at"))
        return ArchiveElement(id, type, actor, repo, payload, public, created_at)

    def to_dict(self) -> dict:
        result: dict = {}
        result["id"] = from_union([from_str, from_none], self.id)
        result["type"] = from_union([from_str, from_none], self.type)
        result["actor"] = from_union(
            [lambda x: to_class(Actor, x), from_none], self.actor
        )
        result["repo"] = from_union([lambda x: to_class(Repo, x), from_none], self.repo)
        result["payload"] = from_union(
            [lambda x: to_class(Payload, x), from_none], self.payload
        )
        result["public"] = from_union([from_bool, from_none], self.public)
        result["created_at"] = from_union(
            [lambda x: x.isoformat(), from_none], self.created_at
        )
        return result


class Archive:
    data: List[ArchiveElement]

    def __init__(self, data: List[ArchiveElement]):
        self.data = data

    @classmethod
    def from_dict_list(cls, data: List[dict]):
        archive_elems = [ArchiveElement.from_dict(d) for d in data]
        return cls(archive_elems)

    @classmethod
    def from_response(cls, resp: requests.Response):
        return cls.from_gzip_bytes(resp.content)

    @classmethod
    def from_gzip_bytes(cls, b: bytes):
        data_str = gzip.decompress(b).decode("utf8")
        data_strs = [s for s in data_str.split("\n") if s]
        del data_str
        json_str = "[" + ", ".join(data_strs) + "]"
        del data_strs
        data = json.loads(json_str)
        return cls.from_dict_list(data)

    def to_dict_list(self) -> List[dict]:
        return [elem.to_dict() for elem in self.data]

    def __add__(self, other):
        cls = self.__class__
        return cls(self.data + other.data)

    def filter(
        self, filters: Sequence[Tuple[str, Union[int, float, str]]]
    ) -> "Archive":
        new_elems = []
        for elem in self.data:
            valid = True
            for (attr, value) in filters:
                getattr_list = attr.split(".")
                filter_value = elem
                for gattr in getattr_list:
                    filter_value = getattr(filter_value, gattr)
                if filter_value != value:
                    valid = False
                    break
            if valid:
                new_elems.append(elem)
        cls = self.__class__
        return cls(new_elems)



import unittest
import os

from gharchive.models import Archive, ArchiveElement

DATA_DIR = os.path.join("tests", "data")
DATA_FILE = os.path.join(DATA_DIR, "2015-01-01-1.json.gz")


class TestCreateModels(unittest.TestCase):
    archive: Archive
    elem: ArchiveElement

    def setUp(self) -> None:
        with open(DATA_FILE, "rb") as f:
            self.archive = Archive.from_gzip_bytes(f.read())
            self.elem = self.archive.data[0]

    def test_create(self):
        assert self.elem.to_dict() == {
            "id": "2489395767",
            "type": "PushEvent",
            "actor": {
                "id": 1310570,
                "login": "soumith",
                "display_login": None,
                "gravatar_id": "",
                "url": "https://api.github.com/users/soumith",
                "avatar_url": "https://avatars.githubusercontent.com/u/1310570?",
            },
            "repo": {
                "id": 28067809,
                "name": "soumith/fbcunn",
                "url": "https://api.github.com/repos/soumith/fbcunn",
            },
            "payload": {
                "ref": "refs/heads/master",
                "ref_type": None,
                "pusher_type": None,
                "push_id": 536752122,
                "size": 4,
                "distinct_size": 4,
                "head": "fa6048ec9b9eeafd12cee5f81324f355e1f2a198",
                "before": "2d06657267b32e0c8e193c617039da200f710195",
                "commits": [
                    {
                        "sha": "dbd68d30ee1f7b60d404553fc1c6226ebb374c8e",
                        "author": {
                            "name": "Soumith Chintala",
                            "email": "88de463b5797707cf3425f85a415c3d869db732b@gmail.com",
                        },
                        "message": "back to old structure, except lua files moved out",
                        "distinct": True,
                        "url": "https://api.github.com/repos/soumith/fbcunn/commits/dbd68d30ee1f7b60d404553fc1c6226ebb374c8e",
                    },
                    {
                        "sha": "5567f9f5a83d7fe3320b18e5b89405e8a5ca77e6",
                        "author": {
                            "name": "Soumith Chintala",
                            "email": "88de463b5797707cf3425f85a415c3d869db732b@gmail.com",
                        },
                        "message": "...",
                        "distinct": True,
                        "url": "https://api.github.com/repos/soumith/fbcunn/commits/5567f9f5a83d7fe3320b18e5b89405e8a5ca77e6",
                    },
                    {
                        "sha": "58a83b277328eca811d3a37cf171b2fc4fcd87af",
                        "author": {
                            "name": "Soumith Chintala",
                            "email": "88de463b5797707cf3425f85a415c3d869db732b@gmail.com",
                        },
                        "message": "...",
                        "distinct": True,
                        "url": "https://api.github.com/repos/soumith/fbcunn/commits/58a83b277328eca811d3a37cf171b2fc4fcd87af",
                    },
                    {
                        "sha": "fa6048ec9b9eeafd12cee5f81324f355e1f2a198",
                        "author": {
                            "name": "Soumith Chintala",
                            "email": "88de463b5797707cf3425f85a415c3d869db732b@gmail.com",
                        },
                        "message": "...",
                        "distinct": True,
                        "url": "https://api.github.com/repos/soumith/fbcunn/commits/fa6048ec9b9eeafd12cee5f81324f355e1f2a198",
                    },
                ],
            },
            "public": True,
            "created_at": "2015-01-01T01:00:00+00:00",
        }


class TestFilterModels(TestCreateModels):
    def test_single_filter(self):
        compare_str = "WatchEvent"
        archive = self.archive.filter([("type", compare_str)])
        for elem in archive.data:
            assert elem.type == compare_str
        assert len(archive.data) > 0

    def test_nested_filter(self):
        compare_str = "https://api.github.com/users/soumith"
        archive = self.archive.filter([("actor.url", compare_str)])
        for elem in archive.data:
            assert elem.actor.url == compare_str
        assert len(archive.data) > 0

    def test_dual_filter(self):
        compare_str_1 = "https://api.github.com/users/jamezb"
        compare_str_2 = "WatchEvent"
        archive = self.archive.filter(
            [("actor.url", compare_str_1), ("type", compare_str_2)]
        )
        for elem in archive.data:
            assert elem.actor.url == compare_str_1
            assert elem.type == compare_str_2
        assert len(archive.data) > 0
